#ifndef CFCGI_LIB_CONTEXT_H
#define CFCGI_LIB_CONTEXT_H

#include "fcgi_packet.h"
#include "cfcgi_request.h"
#include "cfcgi_response.h"
#include "cfcgi_session.h"
#include "simplelog.h"
#if COROUTINE > 1
#include <boost/coroutine2/coroutine.hpp>
using CoroPull = boost::coroutines2::coroutine<FcgiMessagePtr>::pull_type;
using CoroPush = boost::coroutines2::coroutine<FcgiMessagePtr>::push_type;
#else
#include <boost/coroutine/all.hpp>
using CoroPull = boost::coroutines::coroutine<void(FcgiMessagePtr)>::caller_type;
using CoroPush = boost::coroutines::coroutine<void(FcgiMessagePtr)>;
#endif
#include "cfcgi_db.h"
#include "cfcgi_http_client.h"
#include <list>
#include <sys/epoll.h>

#define MK_EVT_DATA(threadid,coid) (((threadid & 0xffff) << 16) | (coid & 0xffff))
#define GET_COID_FROM_EVT(e) (e & 0xffff)
#define GET_THREAD_ID_FROM_EVT(e) (e >> 16)


using CoroPullPtr = std::shared_ptr<CoroPull>;
using CoroPushPtr = std::shared_ptr<CoroPush>;

enum HttpStatus {
    HTTP_OK = 200,
    HTTP_MOVED_PERMANENT = 301,
    HTTP_MOVED_TEMPORARY = 302,
    HTTP_INTERNAL = 500
};

class FcgiContext {
public:
    FcgiContext(const int& id, int pollid, CFcgiSockPtr sk, CoroPull& gco, SimpleLog::Logger::LoggerPtr log,
                const size_t& thid, const std::string& sessdir = "");
    ~FcgiContext();

    // api for handler
    void SessionStart();
    void SessionDestroy();
    void Redirect(const std::string& where);
    CursorPtr GetDb(const std::string& driver, const DbParam& param);
    void Flush();
    bool SaveFile(const std::string& src, const std::string& dest);
    void Exit(int status = HTTP_OK);
    FcgiBaseHttpPtr HTTP();
    FcgiBaseHttpPtr HTTPS();
    const std::string& Query(const std::string& key, const std::string& dft = "");
    const std::string& PostForm(const std::string& key, const std::string& dft = "");
    void SetCookie(const std::string& cookie);
    void SetCookies(const std::vector<std::string>& keys, const std::vector<std::string>& values);
    const std::string& GetCookie(const std::string& key, const std::string& dft = "");
    [[noreturn]] void Die(const std::string& errmsg);
    FcgiValue& GetHeader(const std::string& name);
    void SetHeader(const std::string& name, const std::string& value);
    void LogInfo(const char* fmt, ...);
    void LogErr(const char* fmt, ...);
public:
    // api for module
    void WaitEvent();

    void PollAdd(int sock, unsigned int evt = EPOLLIN);
    void PollDel(int sock, unsigned int evt = EPOLLIN);
    void PollMod(int sock, unsigned int evt);
    void SetRequestId(Uint16 reqid);
    void SetThreadId(const int& id);
    int  GetThreadId();

private:
    void encodeFcgiStdErr(std::string& dest);
    void encodeFcgiStdOut(std::string& dest);
    void encodeFcgiEnd(std::string& dest);
    void sendMsgToWebServer(std::string& msg);
    void onResponse();
    void Filter();
public:
    // data for handler or module
    int          m_id;
    FcgiRequest  Request;
    FcgiResponse Response;
    FcgiSession  Session;
    int          Status;
    std::stringstream& Out;
    std::stringstream& Err;
    SimpleLog::Logger::LoggerPtr Log;
    Uint16       m_requestId;
    bool         m_completed;

private:
    CoroPull&    GetterCo;
    CFcgiSockPtr m_sock;
    bool         m_errored;
    int          m_pollid;
    std::string  m_logprefix;
    std::list<CursorPtr> m_cursors;
    size_t       m_thread;
};

#endif
