#ifndef CFCGI_LIB_DB_H
#define CFCGI_LIB_DB_H
#include <vector>
#include <map>
#include <string>
#include <memory>
#if __cplusplus > 201103L
#include <variant>
using std::variant;
#else
#include <boost/variant.hpp>
using boost::variant;
#endif
#include <initializer_list>

#define DB_IDLE_MAX_SEC 180

#define DB_POOL_MIN_SIZE 20

class FcgiContext;
class Db;
typedef std::shared_ptr<Db> DbPtr;
class DbMgr;

typedef variant<int, double, std::string> BindType;
typedef std::initializer_list<BindType> BindList;
// typedef std::variant<int*, double*, std::string*> ColumnType;
typedef std::initializer_list<std::string*> RowType;

struct DbParam {
    std::string host;
    std::string user;
    std::string password;
    std::string dbname;
    std::string charset;
    unsigned short port;
};

class Cursor {
    friend class Db;
public:
    Cursor(const size_t& id, FcgiContext* ctx, Db* owner);
    virtual ~Cursor() {}

    virtual bool TransBegin() = 0;
    virtual bool Commit() = 0;
    virtual bool Rollback() = 0;
    virtual bool Connect() = 0;
    virtual bool Ping() = 0;
    virtual bool Disconnect() = 0;
    virtual bool Query(const char* sql, ...) = 0;
    virtual bool QueryBind(const std::string& sql, const BindList& plist) = 0;
    virtual bool FetchRow(RowType row) = 0;
    virtual size_t FieldCount() = 0;
    virtual bool NextResult() = 0;
    virtual void Close() {}
protected:
    Db* m_owner;
    FcgiContext* m_ctx;
    size_t m_id;
};

typedef std::shared_ptr<Cursor> CursorPtr;

class Db {
public:
    typedef std::vector<CursorPtr> PoolItem;

    friend class DbMgr;
    Db(const DbParam& param);
    virtual ~Db();

    virtual CursorPtr GetCursor(FcgiContext& ctx);
    virtual void RelCursor(const size_t& id);
    inline DbParam& GetParam() { return m_param; }
protected:
    virtual CursorPtr CreateCursor(const size_t& id, FcgiContext& ctx) = 0;
protected:
    DbParam m_param;
    PoolItem m_cursors;
    PoolItem m_freeList;
};

typedef std::shared_ptr<Db> DbPtr;

class Driver {
public:
    Driver() {}
    virtual ~Driver();
    virtual DbPtr Create(const DbParam& param) = 0;
};

typedef std::shared_ptr<Driver> DriverPtr;
typedef std::vector<DbPtr> DbPool;
typedef std::map<std::string, DriverPtr> DriverMap;
typedef std::map<std::string, DbPool> DbPoolMap;

class DbMgr {
private:
    DbMgr() {}
public:
    ~DbMgr() {}
    static void RegisterDb(const std::string& driverName, DriverPtr driver);

    static DbPtr GetDb(const std::string& driverName, const DbParam&Param);
private:
    static DriverMap m_drivers;
    static DbPoolMap m_pools;
};

#endif // CFCGI_LIB_DB_H
