#ifndef CFCGI_LIB_OBJECT_H
#define CFCGI_LIB_OBJECT_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <initializer_list>
#include <tuple>
#if __cplusplus > 201103L
#include <variant>
using std::variant;
using std::tuple;
#else
#include <boost/variant.hpp>
using boost::variant;
extern const size_t INT_INDEX;
extern const size_t DOUBLE_INDEX;
extern const size_t STRING_INDEX;
#endif
#include <memory>
#include <string.h>

typedef std::vector<int> intarray;
typedef std::vector<std::string> strarray;
typedef std::vector<double> floatarray;

class FcgiObject;
typedef std::vector<FcgiObject*> objarray;

class ObjectValue {
public:
    typedef variant<int, double, std::string> InputDataType;
    typedef std::tuple<std::string, InputDataType> KeyValueType;
    typedef std::tuple<std::string, std::initializer_list<std::string>> KeyStrList;
    typedef std::tuple<std::string, std::initializer_list<int>> KeyIntList;
    typedef std::tuple<std::string, std::initializer_list<double>> KeyFloatList;
    typedef std::tuple<std::string, std::initializer_list<KeyValueType>> KeyObjList;
    typedef std::vector<ObjectValue*> valarray;

    enum VarLoc {
        INT_LOC = 0,
        FLOAT_LOC = 1,
        STRING_LOC = 2
    };

    enum ValueType {
        OV_INVALID,
        OV_INT,
        OV_DOUBLE,
        OV_STRING,
        OV_OBJECT,
        OV_INT_ARR,
        OV_DOUBLE_ARR,
        OV_STRING_ARR,
        OV_OBJECT_ARR,
        // OV_ARRAY_ARR,

        N_OVS
    };

    ObjectValue(ValueType t, void *v);
    ObjectValue(const int& v);
    ObjectValue(const double& v);
    ObjectValue(const std::string& v);
    ObjectValue(const intarray& v);
    ObjectValue(const strarray& v);
    ObjectValue(FcgiObject* v);
    ObjectValue(const floatarray& v);
    ObjectValue(const objarray& v);

    virtual ~ObjectValue();

    void Free();

    bool isValid();

    operator int() const;
    operator double() const;
    operator std::string&() const;
    operator FcgiObject* () const;

    operator intarray() const;
    operator floatarray() const;
    operator strarray() const;
    operator objarray() const;

    ObjectValue& operator[](const int& index);
    ObjectValue& operator[](const std::string& key);

    ObjectValue& operator=(const int& v);
    ObjectValue& operator=(const double& d);
    ObjectValue& operator=(const std::string& v);
    ObjectValue& operator=(FcgiObject* v);
    ObjectValue& operator=(const intarray& v);
    ObjectValue& operator=(const floatarray& v);
    ObjectValue& operator=(const strarray& v);
    ObjectValue& operator=(const objarray& v);

    friend bool operator==(const ObjectValue& lh, const int& rh);
    friend bool operator==(const int& lh, const ObjectValue& rh);

    friend bool operator==(const std::string& lh, const ObjectValue& rh);
    friend bool operator==(const ObjectValue& lh, const std::string& rh);

    friend bool operator==(const double& lh, const ObjectValue& rh);
    friend bool operator==(const ObjectValue& lh, const double& rh);

    // array methods, used to add data to array, return current object.
    ObjectValue& operator-(const int& v);
    ObjectValue& operator-(const std::initializer_list<int>& v);
    ObjectValue& operator-(const std::string& v);
    ObjectValue& operator-(const std::initializer_list<std::string>& v);
    ObjectValue& operator-(const double& v);
    ObjectValue& operator-(const std::initializer_list<double>& v);
    FcgiObject& operator-(const ObjectValue::KeyValueType& kv);
    FcgiObject& operator-(const ObjectValue::KeyStrList& kvs);
    FcgiObject& operator-(const ObjectValue::KeyIntList& kvs);
    FcgiObject& operator-(const ObjectValue::KeyFloatList& kvs);

    // methods for add key, value
    // add a key, return the new object
    FcgiObject& operator+(const std::string& key);

    FcgiObject& GetObject();
    friend std::ostream& operator<<(std::ostream& o, const ObjectValue& v);
    friend std::istream& operator>>(std::istream& o, ObjectValue& v);
    friend std::ofstream& operator<<(std::ofstream& o, const ObjectValue& v);
    friend std::ifstream& operator>>(std::ifstream& o, ObjectValue& v);

    ValueType   m_type;
    void*       m_data;

private:
    void setInt(const int& v);
    void setFloat(const double& v);
    void setString(const std::string& v);
    void setObject(const FcgiObject* v);
    void setIntArr(const intarray& v);
    void setStrArr(const strarray& v);
    void setFloatArr(const floatarray& v);
    void setObjectArr(const objarray& v);
};

struct StrLess {
    bool operator()(const std::string& lh, const std::string& rh) const {
        return strcmp(lh.c_str(), rh.c_str()) < 0;
    }
};

class FcgiObject {
public:
    typedef std::map<std::string, ObjectValue*, StrLess> ObjKeyMap;
    typedef std::initializer_list<ObjectValue::KeyValueType> SimpleJson;

    FcgiObject();
    FcgiObject(const SimpleJson& v);
    ~FcgiObject();
    void Reset();

    ObjectValue& operator[](const std::string& key);

    // methods for add key, value
    // add a key return the new object
    FcgiObject& operator+(const std::string& key);

    // add key values, return current object
    FcgiObject& operator-(const ObjectValue::KeyValueType& kv);
    FcgiObject& operator-(const ObjectValue::KeyStrList& kvs);
    FcgiObject& operator-(const ObjectValue::KeyIntList& kvs);
    FcgiObject& operator-(const ObjectValue::KeyFloatList& kvs);
    FcgiObject& operator-(const ObjectValue::KeyObjList& kvs);

    friend std::ostream& operator<<(std::ostream& o, const FcgiObject& v);
    friend std::istream& operator>>(std::istream& o, FcgiObject& v);

    friend std::ofstream& operator<<(std::ofstream& o, const FcgiObject& v);
    friend std::ifstream& operator>>(std::ifstream& o, FcgiObject& v);
    ObjKeyMap& GetValues();
protected:
    ObjKeyMap m_value;
};

typedef std::shared_ptr<FcgiObject> FcgiObjectPtr;

extern std::string NullString;
extern intarray NullIntArray;
extern strarray NullStrArray;
extern floatarray NullFloatArray;
extern objarray NullObjArray;
extern ObjectValue NullObjValue;

const ObjectValue::KeyIntList& _F(const ObjectValue::KeyIntList& x);
// const ObjectValue::KeyFloatList& _F(const ObjectValue::KeyFloatList& x);
const ObjectValue::KeyStrList& _F(const ObjectValue::KeyStrList& x);
const ObjectValue::KeyObjList& _F(const ObjectValue::KeyObjList& x);

typedef ObjectValue::KeyValueType _K;
void skipSpace(std::istream& o);

#endif // CFCGI_LIB_OBJECT_H
