#ifndef CFCGISOCK_H
#define CFCGISOCK_H

#include <tuple>
#include <string>
#include <netinet/in.h>
#include <sys/un.h>
#include <memory>

typedef const char* error;
typedef std::tuple<std::string, unsigned short> Address;
typedef std::tuple<std::string&, unsigned short&> AddressRef;

class CFcgiSock
{
public:
    CFcgiSock();
    virtual ~CFcgiSock();
    virtual error Init(const std::string& addr) = 0;
    virtual error Init(char addr[], const size_t& size, unsigned short port) { return nullptr; }
    virtual error Init(struct sockaddr* addr, const size_t& size) { return nullptr; }
    virtual error Listen(int backlog = 20);
    virtual error Accept(CFcgiSock& sock);
    virtual error Connect();
    virtual int getSock() = 0;
    virtual std::tuple<int, error> Send(const std::string& data) = 0;
    virtual std::tuple<int, error> Recv(std::string& data) = 0;
    virtual std::tuple<int, error> Recv(char* data, int rlen) = 0;
    virtual void ReadReady(bool ready) { m_canRead = ready; }
    virtual bool ReadReady() { return m_canRead; }
    virtual void WriteReady(bool ready) { m_canWrite = ready; }
    virtual bool WriteReady() { return m_canWrite; }
    virtual void Close() = 0;
protected:
    std::tuple<int, error> SendData(int sock, const std::string& data);
    std::tuple<int, error> RecvData(int sock, std::string& data);
    std::tuple<int, error> RecvData(int sock, char* data, int rlen);
protected:
    bool m_canRead;
    bool m_canWrite;
};
using CFcgiSockPtr = std::shared_ptr<CFcgiSock>;

class CFcgiNetSock : public CFcgiSock {
public:
    CFcgiNetSock(int sock = 0) : m_sock(sock) {}
    virtual ~CFcgiNetSock() {}

    virtual error Init(const std::string& addr);
    virtual error Init(char addr[], const size_t& size, unsigned short port);
    virtual error Init(struct sockaddr* addr, const size_t& size);
    virtual error Listen(int backlog = 20);
    virtual error Accept(CFcgiSock& sock);
    virtual error Connect();
    virtual int getSock() { return m_sock; }
    virtual std::tuple<int, error> Send(const std::string& data);
    virtual std::tuple<int, error> Recv(std::string& data);
    virtual std::tuple<int, error> Recv(char* data, int rlen);
    virtual void Close();

    Address parseAddr(const std::string& addr);

    int GetSocket() { return m_sock; }
    void SetSocket(int sock) { m_sock = sock; }
    error Connected();

    void GetSockAddr(struct sockaddr_in& addr) { addr = m_addr; }
    void SetSockAddr(struct sockaddr_in& addr) { m_addr = addr; }
private:
    error initSock();

private:
    int m_sock;
    struct sockaddr_in m_addr;
};
using CFcgiNetSockPtr = std::shared_ptr<CFcgiNetSock>;

class CFcgiUnixSock : public CFcgiSock {
public:
  CFcgiUnixSock() {}
  virtual ~CFcgiUnixSock() {}
  virtual error Init(const std::string& addr);
  virtual error Listen(int backlog = 20);
  virtual error Accept(CFcgiSock& sock);
  virtual error Connect();
  virtual int getSock() { return m_sock; }
  virtual std::tuple<int, error> Send(const std::string& data);
  virtual std::tuple<int, error> Recv(std::string& data);
  virtual std::tuple<int, error> Recv(char* data, int rlen);
  virtual void Close();
  Address parseAddr(const std::string& path);
private:
  error initSock();
 private:
  std::string m_path;
  int m_sock;
  struct sockaddr_un m_addr;
};
using CFcgiUnixSockPtr = std::shared_ptr<CFcgiUnixSock>;

class CFcgiPipe : public CFcgiSock {
public:
  enum PipeType {
    CFCGI_NAMED_FIFO,
    CFCGI_PIPE
  };
  CFcgiPipe() {}
  virtual ~CFcgiPipe() {}
  virtual error Init(const std::string& addr = "");
  virtual int getSock() { return m_pipe[0]; }
  virtual std::tuple<int, error> Send(const std::string& data);
  virtual std::tuple<int, error> Recv(std::string& data);
  virtual std::tuple<int, error> Recv(char* data, int rlen);
  virtual void Close();
private:
  std::string m_path;
  PipeType    m_pipeType;
  int         m_pipe[2];
};
using CFcgiPipePtr = std::shared_ptr<CFcgiPipe>;

#endif // CFCGISOCK_H
