#ifndef SIMPLELOG_H
#define SIMPLELOG_H

#include <string>
#include <list>
#include <memory>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>
#include <map>

#ifdef USE_MULTI_THREAD
#include <mutex>
#endif

namespace SimpleLog {

const int MAX_LOG_BLOCK = 102400;
const int MAX_CACHE_BLOCKS = 10;
const int FLUSH_INTERVAL = 5; // 5 seconds

enum LogLevel {
    FATAL,
    ALERT,
    ERROR,
    WARN,
    INFO,
    DEBUG,
    N_LEVELS
};

// cacheBlock
struct cacheBlock {
    char m_block[MAX_LOG_BLOCK];
    int  m_pos;

    int Left() { return MAX_LOG_BLOCK - m_pos; }
};

typedef std::shared_ptr<cacheBlock> cacheBlockPtr;
class logCache {
public:
    typedef std::list<cacheBlockPtr> CachePool;
    typedef CachePool::iterator BlockType;
    logCache();
    ~logCache();

    char *getBuffer(int& sz);
    void putBack(const int& sz);

    logCache& operator<<(const char* s);
    logCache& operator<<(const std::string& s);
    logCache& operator<<(const int& i);
    logCache& operator<<(const long long& l);

    void Clear();

private:
    void addBlock();
    void writeCache(void *data, const int& len);
public:
    CachePool m_caches;
    BlockType m_curBlock;
};

#ifdef USE_MULTI_THREAD
class ScopeLock {
public:
    ScopeLock(std::mutex& lock);
    ~ScopeLock();
private:
    std::mutex& m_lock;
};
#endif

class Logger {
private:
    Logger(const std::string& name, const std::string& suff = "");
public:
    typedef std::shared_ptr<Logger> LoggerPtr;
    ~Logger();

    static void SetLogDir(const std::string& logdir);
    static LoggerPtr GetLogger(const std::string& name, const std::string& suff = "");
    static void FlushAll();

    void SetCache(bool cached);
    bool GetCache();
    void SetLevel(LogLevel level);
    LogLevel GetLevel();
    void Flush(const time_t& t = 0);

    void Debug(const char* fmt, ...);
    void Info(const char* fmt, ...);
    void Warn(const char* fmt, ...);
    void Err(const char* fmt, ...);
    void Alert(const char* fmt, ...);
    void Fatal(const char* fmt, ...);
    void log(LogLevel level, const std::string& prefix, const char* fmt, va_list va);
private:
    void rotate(time_t& logtime, const char *slogtime);
    bool needRotate(time_t& logtime);
    void log(const std::string& level, int sz, const char* buf);
    bool renameFile(const std::string& src, const std::string& dest);
    std::string makeFilePath(const std::string& partone, const std::string& part2);
private:
    static std::map<std::string, LoggerPtr> m_instances;
    LogLevel m_level;
    bool     m_cached;
    std::string m_name;
    std::string m_nameSuffix;
    static std::string m_dir;
    std::string m_path;
    FILE*   m_file;
    time_t m_roateTime;
    time_t m_lastFlushTime;
    logCache m_data;
#ifdef USE_MULTI_THREAD
    std::mutex m_lock;
#endif
}; // Logger
}; // SimpleLog

#endif // SIMPLELOG_H
